// =====================================================================================================================================================================================================
// (c) 2022 Lynn Hansen, KU7Q															                                                                                                               |
// This Source Code Form is subject to the terms of the GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. A copy of this license can be found here: https://choosealicense.com/licenses/gpl-3.0/|
// =====================================================================================================================================================================================================



// THIS MODULE CONTAINS ALL THE CODE TO INTERFACE TO FLEX RADIOS


void Flex_Initialize()
{
	/*
	NOTE:
	The Flex radio only updates the transmit rfpower and tunepower when it actually transmits
	To set the power or tune power you have to enter the new power levels then transmit (or start the atu) otherwise they will reset to the old levels on the next transmit subscription update
	*/

	//we're starting (or restarting) the flex comm link - list of active slices
	//If no active slices, create one
	//Subscribe to slice and Tx

	//don't connect to flex if Link is connected and we're in Sync: As Remote mode
	if (gLinkMiniA && gSyncMode == SYNC_MODE_RMT && (gBLEConnected || gUDPConnected))
	{
		return;
	}

	String cmd = "";
	String reply = "";

	static bool alreadyHr = false;

	uint32_t flexStartTm = millis();

	if (!gFlexInit)
	{
		gFreqRadioPrev = 0; //reset so we send new freq & mode on new slice connect
		gRModePrev = 0xff;
	}

	gFlexInit = false; //block async tx polls - goes true after all initization is done

	if (gIPAdrs.indexOf(".") == -1)
	{
		PrtStatus("Invalid Flex IP Adrs", true);
		gTCPCon = TCP_CON_OFF;
		PrtOptions();
		return;
	}

	if (alreadyHr || !gWiFi || !gTCPCon)
	{
		return;
	}
	alreadyHr = true;

	//uint8_t tryCtr = 0; //try 4 times to connect then fail
	char adrs[20];
		
	gIPAdrs.toCharArray(adrs, gIPAdrs.length() + 1);
	const char* fAdrs = adrs; //must be a const
	//Serial.println("Flex adrs= " + gIPAdrs + ", " + String(adrs) + ", " + String(fAdrs));

	tcp.stop();
	delay(100);
	tcp = server.available();	
	
	tcp.setTimeout(10); //in seconds
	delay(500);

	if (!tcp.connected())
	{		
		hmiTagLineRefresh = 0;
		PrtStatus("Connect: " + gIPAdrs, true);
		PrtOptions();
		int fPort = gFlexPort.toInt();
		tcp.connect(fAdrs, fPort);		
		delay(500);
		if (!tcp.connected())
		{			
BadConn:	hmiTagLineRefresh = 0;
			PrtStatus("Flex Connection FAILED", true);
			SerialOut("Failed to connect to " + String(fAdrs), true);
			tcp.stop();
			gTCPCon = TCP_CON_OFF;
			PrtOptions();
			gFlexInit = false;
			alreadyHr = false;
			return;
		}		
		gTCPCon = FLEX_CON_CONNECTING; //connecting, not bound
		gFlexCmdNum = 1; //restart counter for this session	 
		gFlexHandle = ""; //reset - get new handle in first reply msg
		gFlexHandleTx = "";
		gFlexClientID = "";
		//PrtStatus("Connected to Flex: " + gFlexSliceLbl[gFlexSlice], true);		
		flexStartTm = millis();
		while (5000 > millis() - flexStartTm)
		{
			//wait up to 5 seconds for initial data dump to get handle
			if (tcp.available())
			{
#ifdef DEBUG_FLEX_API
				SerialOut("Flex Init time= " + String(millis() - flexStartTm), true);
#endif
				break;
			}
		}
		delay(1000); //wait for the entire initialization string to show up
		Flex_ChkRx(); //get the v# and handle		
		
		if (!gFlexHandle.length())
		{
			goto BadConn; //failed to get handle
		}
	}

	Tx2WiFi(WIFI_FLEX, "|ping"); //get initial connect dump
	Tx2WiFi(WIFI_FLEX, "|slice list");	
	String id = gFlexHandle.substring(1, gFlexHandle.length() - 1); //get handle ID#	

	gRIT = 0;
	SetRIT(); //make sure this is off
	PrtRIT();

	//subscibe to slice, tx, and atu data	
	Tx2WiFi(WIFI_FLEX, "|sub radio all");
	Tx2WiFi(WIFI_FLEX, "|sub slice " + String(gFlexSlice)); //all"); 
	Tx2WiFi(WIFI_FLEX, "|sub tx all");
	Tx2WiFi(WIFI_FLEX, "|sub atu all");
	Tx2WiFi(WIFI_FLEX, "|sub client all"); //get client ID

	Tx2WiFi(WIFI_FLEX, "|audio client " + id + " slice " + String(gFlexSlice) + " gain " + String(gFlexVol));

	//wait for 2 sec for replies to above	
	for (int i = 0; i < 200; i++)
	{
		if (tcp.available())
		{
			Flex_ChkRx();
		}
		delay(10);
	}

	//set DSP filters to values saved in HMI eeprom - radio doesn't report filter settings on startup??

	//zero these out and let mini update from flex
	gFlexWNB = 0;
	gFlexNB = 0;
	gFlexNR = 0;
	gFlexAPF = 0;
	gFlexANF = 0;

	gFlexInit = true; //start async polling
	alreadyHr = false;
}

void Flex_Final_Init()
{
	//this is called after we have done the pre-initialization and created the slice we need
	Tx2WiFi(WIFI_FLEX, "|keepalive disable"); //if enabled you must ping every second - radio closes connection after 15 seconds - we ping every 10 seconds and fail after 60 seconds

	Tx2WiFi(WIFI_FLEX, "|slice s " + String(gFlexSlice) + " active=1 tx=1 txant=ANT" + String(gFlexAnt));	
}

String Flex_CalcFreq(uint32_t freq)
{
	//convert freq to floating for Flex command - can't do it directly because of float rounding errors
	uint32_t frq = freq;
	uint32_t val = 0;
	String f = "";
	if (frq > 99999999)
	{
		val = frq / 100000000;
		f = String(val);
		frq -= val * 100000000;
	}
	if (frq > 9999999)
	{
		val = frq / 10000000;
		f += String(val);
		frq -= val * 10000000;
	}
	else if (freq > 99999999)
	{
		f += "0";
	}
	if (frq > 999999)
	{
		val = frq / 1000000;
		f += String(val);
		frq -= val * 1000000;
	}
	else if (freq > 9999999)
	{
		f += "0";
	}
	f += "."; //add the decimal point
	if (frq > 99999)
	{
		val = frq / 100000;
		f += String(val);
		frq -= val * 100000;
	}
	else
	{
		f += "0";
	}
	if (frq > 9999)
	{
		val = frq / 10000;
		f += String(val);
		frq -= val * 10000;
	}
	else
	{
		f += "0";
	}
	if (frq > 999)
	{
		val = frq / 1000;
		f += String(val);
		frq -= val * 1000;
	}
	else
	{
		f += "0";
	}
	if (frq > 99)
	{
		val = frq / 100;
		f += String(val);
		frq -= val * 100;
	}
	else
	{
		f += "0";
	}
	if (frq > 9)
	{
		val = frq / 10;
		f += String(val);
		frq -= val * 10;
	}
	else
	{
		f += "0";
	}
	f += String(frq);
	return f;
}

String Flex_CalcMode()
{
	//calculates the correct flex mode from the ctr2 modes
	String md = "";
	switch (gRMode)
	{
	case RADIO_CW:
		md = "cw";
		break;
	case RADIO_CW_R:
		md = "cw";
		break;
	case RADIO_LSB:
		md = "lsb";
		break;
	case RADIO_USB:
		md = "usb";
		break;
	case RADIO_FM_N:
		md = "nfm"; //3khz bw
		break;
	case RADIO_FM_W:
		md = "fm"; //5khz bw
		break;
	case RADIO_AM:
		md = "am";
		break;
	case RADIO_DIG_L:
		md = "digl";
		break;
	case RADIO_DIG_H:
		md = "digu";
		break;
	}
	return md;
}


void Flex_ChkRx()
{
	//Code from CTR2-Mini
	static bool alreadyHr = false; //prevent recursion

	if (alreadyHr == true)
	{
		return;
	}
	alreadyHr = true;

	if (tcp.available())
	{
		gFlexLastReply = millis(); //to detect Flex comm failure after 5 seconds		

		//chk for handle
		String txt;
		byte chr[1];
	TryAgn: txt = "";
		while (tcp.available())
		{
			tcp.readBytes(chr, 1);
			if (chr[0] != 10 && chr[0] != 13)
			{
				txt += char(chr[0]);
			}
			else
			{
				goto GotOne;
			}
		}
		if (!txt.length())
		{
			if (!tcp.available())
			{
				alreadyHr = false;
				return;
			}
			goto TryAgn;
		}

	GotOne:

		if (gFlexCmdNum == 1)
		{
			//SerialOut(txt, true);
			if (txt.indexOf('H') == 0)
			{
				int ptr = txt.indexOf("H") + 1;
				int end = txt.indexOf(10, ptr); //look for lf			
				//SerialOut("ptr= " + String(ptr) + ", end= " + String(end), true);
				gFlexHandle = "S" + txt.substring(ptr, end) + "|"; //form full handle to look for
				gFlexHandle.trim();
				//remove leading 0's - flex doesn't send them in 'S' handle '|'
				while (gFlexHandle.charAt(1) == '0')
				{
					gFlexHandle.remove(1, 1);
				}
#ifdef DEBUG_FLEX_API
				SerialOut("\n\r ===> handle= " + gFlexHandle + "\n\r", true);
#endif
			}
		}
		else if (txt.length() && txt.indexOf("R2|0|") > -1)
		{
			//SerialOut(txt, true);
			Chk4ExistingSlice(txt);
		}

		Flex_Parse_Rx(txt);

		if (!gFlexClientID.length())
		{
			//chk to see if we're past Rx record 11 - if we are and haven't received a clientID register as gui
			//SerialOut("No Client ID: " + txt, true);
			if (txt.indexOf("R10|") > -1)
			{
				Tx2WiFi(WIFI_FLEX, "|client gui");
			}
		}
		goto TryAgn;
	}
	alreadyHr = false;
	
}

void Chk4ExistingSlice(String txt)
{
	//NEW CTR2 CODE
	//checks for an existing slice in the R2|0| reply - if one doesn't exist, create one
	//we always start with a ping first then get the slice list			
	//only do this on R2|		

#ifdef DEBUG_FLEX_API
	SerialOut("\n\r Flex slice reply= " + txt, true);
	SerialOut("Looking for: " + String(gFlexSlice), true);
#endif
	if (txt.length() < 5)
	{
		txt = "     ";
	}
	else
	{
		txt = txt.substring(5, txt.length());
	}

	if (txt.indexOf(String(gFlexSlice)) < 0)
	{
		//gFlexSlc doesn't exist, create the number of slices it take to get to the slice we want, space them 1khz a part
		String newSlices = ""; //save new slices so we can delete them if they're not the active one
		for (uint8_t i = 0; i < gFlexSlice + 1; i++)
		{
			//SerialOut("chking for " + String(i) + ", " + String(txt.indexOf(String(i))), true);
			if (txt.indexOf(String(i)) < 0)
			{
				newSlices += String(i);
				String frq = Flex_CalcFreq(gFreqRadio + ((i + 1) * 1000)); //start new slice 1khz away
				String md = Flex_CalcMode();

				//SerialOut("\r\n Free Panadapters---> " + String(gFlexFreePanadapters) + "\r\n", true);
				if (gFlexFreePanadapters && gFlexPan)
				{
					//try to create a new panadapter if there's any left
					Tx2WiFi(WIFI_FLEX, "|display pan c freq=" + frq + " ant=ANT" + String(gFlexAnt));
					delay(500);
					if (gFlexFreeSlices)
					{
						Tx2WiFi(WIFI_FLEX, "|slice create freq=" + frq + " pan=" + String(gFlexSlice + 1) + " ant=ANT" + String(gFlexAnt) + " mode=" + md);
						gFlexFreeSlices--; //dec 
					}
				}
				else
				{
					//just try to create the slice
					if (gFlexFreeSlices)
					{
						Tx2WiFi(WIFI_FLEX, "|slice create freq=" + frq + " ant=ANT" + String(gFlexAnt) + " mode=" + md);
						gFlexFreeSlices--; //dec 
					}
				}
				delay(250);
			}
		}
		//chk to see if we created more than one slice, if so, deleted everything but the last one
		if (newSlices.length() > 1)
		{
			for (uint8_t i = 0; i < newSlices.length() - 1; i++)
			{
				Tx2WiFi(WIFI_FLEX, "|slice r " + newSlices.substring(0, 1));
				//SerialOut("Deleting slice " + newSlices.substring(0, 1), true);
			}
		}
	}
	gFlexPan = false; //reset it till next time
	Flex_Final_Init(); //finish setting up the Flex once we have the slice created - this will set the active slice to gFreqRadio
	

}

void Flex_Parse_Rx(String& txt)
{
	String data = ""; //string returned from Flex_Extract()
	uint16_t ptr = 0;
	bool updated = false; //if true and we're on the Flex menu, update the values
	static uint32_t lastTuneMsg = millis(); //wait 10 sec after tune message to check if tx freq has changed

	//parse txt for '=' pair - if found, update variable associated with it. Clear txt if '=' pair found
	if (txt.length() == 0)
	{
		txt = " "; //make sure there's something in it or we might lock up proc
	}

	//set the static slice0 value - true when last |slice frame is for slice 0
#ifdef DEBUG_FLEX_API	
	SerialOut("\n\r==================================================================", true);
	SerialOut("Rx txt:\n\r" + txt, true);
	SerialOut("==================================================================\n\r", true);
#endif

	if (txt.indexOf("|slice " + String(gFlexSlice) + " active=0") > -1)
	{
		//another connection just went active, turn gTxMode off
		gTxMode = RADIO_TX_OFF;
		SendASCCmd(ASC_CMD_TXMODE, 0, 0);
	}

	if (txt.indexOf(gFlexHandle) > -1 || txt.indexOf("|slice " + String(gFlexSlice) + " ") > -1)
	{

		data = Flex_Extract("slices=", txt);
		if (data.length())
		{
			gFlexFreeSlices = data.toInt();
		}

		data = Flex_Extract(" tx_slice_mode=", txt);
		if (data.length())
		{
			goto SetMode;
		}

		data = Flex_Extract("|slice " + String(gFlexSlice) + " mode=", txt);
		if (data.length())
		{
SetMode:	if (data == "CW")
			{
				gRMode = RADIO_CW;
			}
			else if (data == "LSB")
			{
				gRMode = RADIO_LSB;
			}
			else if (data == "USB")
			{
				gRMode = RADIO_USB;
			}
			else if (data == "AM")
			{
				gRMode = RADIO_AM;
			}
			else if (data == "NFM")
			{
				gRMode = RADIO_FM_N;
			}
			else if (data == "FM")
			{
				gRMode = RADIO_FM_W;
			}
			else if (data == "DIGL")
			{
				gRMode = RADIO_DIG_L;
			}
			else if (data == "DIGU")
			{
				gRMode = RADIO_DIG_H;
			}
			updated = true;
			gRModePrev = gRMode; //so we don't send it back to radio
			gVFO_Mode[gVFO] = gRMode;

			PrtRMode(true); //print it here since we can't detect the change in Loop() because we set Prev too
		}
		data = Flex_Extract(" wnb=", txt);
		if (data.length())
		{
			//SerialOut("==> wnb= " + data, true);
			if (data.toInt() == 0)
			{
				gFlexWNB = 0;
			}
			updated = true;
		}
		data = Flex_Extract(" nb=", txt);
		if (data.length())
		{
			//update noise blanker status
			//SerialOut("==> nb= " + data, true);
			if (data.toInt() == 0)
			{
				gFlexNB = 0;
			}
			updated = true;
		}
		data = Flex_Extract(" nr=", txt);
		if (data.length())
		{
			//update noise blanker status
			//SerialOut("==> nr= " + data, true);
			if (data.toInt() == 0)
			{
				gFlexNR = 0;
			}
			updated = true;
		}

		//use auto peaking filter for cw modes
		data = Flex_Extract("apf=", txt);
		if (data.length())
		{
			//SerialOut("==> apf= " + data, true);
			if (data.toInt() == 0)
			{
				gFlexAPF = 0;
			}
			updated = true;
		}

		//use auto notch filter for all other modes
		data = Flex_Extract("anf=", txt);
		if (data.length())
		{
			//update noise blanker status
			//SerialOut("==> anf= " + data, true);
			if (data.toInt() == 0)
			{
				gFlexANF = 0;
			}
			updated = true;
		}

		data = Flex_Extract(" wnb_level=", txt);
		if (data.length())
		{
			//update noise blanker status			
			gFlexWNB = data.toInt();
			updated = true;
		}
		data = Flex_Extract(" nb_level=", txt);
		if (data.length())
		{
			//update noise blanker status
			//SerialOut("==> nb= " + data, true);
			gFlexNB = data.toInt();
			updated = true;
		}
		data = Flex_Extract(" nr_level=", txt);
		if (data.length())
		{
			//update noise blanker status
			//SerialOut("==> nr= " + data, true);
			gFlexNR = data.toInt();
			updated = true;
		}

		//use auto peaking filter for CW
		data = Flex_Extract("apf_level=", txt);
		if (data.length())
		{
			//update auto peaking filter status							
			gFlexAPF = data.toInt();
			updated = true;
		}

		//use auto notch for other modes
		data = Flex_Extract("anf_level=", txt);
		if (data.length())
		{
			//update auto notch filter status				
			gFlexANF = data.toInt();
			//SerialOut("....anf -- gFlexANF= " + String(gFlexANF), true);		
			updated = true;
		}

		data = Flex_Extract(" audio_level=", txt);
		if (data.length())
		{
			//get slice audio gain
			ptr = gFlexVol;
			gFlexVol = data.toInt();
			if (ptr != gFlexVol)
			{
				gSaveRadio = true;
			}
			//SerialOut("--> " + txt + ", Audio Gain = " + data, true);
			updated = true;
		}

		//chk freq first incase we zero it out if rit is off
		data = Flex_Extract(" rit_freq=", txt);
		if (data.length())
		{
			int temp = data.toInt();			
			if (temp != gRIT)
			{
				gRIT = temp;
				PrtRIT();
			}
		}

		data = Flex_Extract(" rit_on=", txt);
		if (data.length() && txt.indexOf('R') != 0)
		{			
			if (data == "0")
			{
				gRIT = 0;				
			}
			PrtRIT();
		}
	}

	//update the Flex menu if anything has changed and it's visible
	if (gMnu == MNU_FLEX && updated)
	{
		CreateMnu("", false);
	}

	//client info
	if (!gFlexHandleTx.length())
	{		
		if (txt.indexOf("|client ") > -1 && txt.indexOf(" client_id=") > -1)
		{
			//get tx handle if we're connecting as the GUI
			ptr = txt.indexOf("|client ");
			if (txt.indexOf("0x") == ptr + 8)
			{				
				ptr += 8; //start at 0x
				gFlexHandleTx = txt.substring(ptr, ptr + 10);
				gFlexHandleTx.trim();
				data = Flex_Extract(" client_id=", txt);
				if (data.length() && data.length() > 30 && data.indexOf('-') > -1)
				{
					if (data.indexOf('S') > -1)
					{
						data = data.substring(0, data.indexOf('S'));
					}
					gFlexClientID = data;
					gFlexClientID.trim();
				}
				data = Flex_Extract(" program=", txt);
				data.toUpperCase();				
				if (data.length() && data.indexOf("SMARTSDR") >-1)
				{
					PrtStatus("Bound as Non-GUI Client", true);
					Tx2WiFi(WIFI_FLEX, "|client bind client_id=" + gFlexClientID);					
					gTCPCon = FLEX_CON_NON_GUI; 
				}
				else
				{
					//already sent client gui command in Flex_RxChk() after not receiving client data from sub client all request
					PrtStatus("Connected as GUI Client", true);										
					gTCPCon = FLEX_CON_GUI;
				}
				PrtOptions();
#ifdef DEBUG_FLEX_API
				SerialOut("\n\n\r++++++++++++++++++++++++++++++++++++++++++++++++++", true);
				SerialOut(data, true);
				SerialOut("client_id= " + gFlexClientID + ", client_handle= " + gFlexHandleTx, true);
				SerialOut("++++++++++++++++++++++++++++++++++++++++++++++++++\n\n\r", true);
#endif
			}
		}					
	}	
	
	//rx info

	data = Flex_Extract(" panadapters=", txt);
	if (data.length())
	{
		//this is the # of free panadapters
		gFlexFreePanadapters = data.toInt();
		//SerialOut("\r\n # of panadapters= " + String(gFlexFreePanadapters) + "\r\n", true);
	}
	
	data = Flex_Extract("|slice " + String(gFlexSlice) + " RF_frequency=", txt);
	if (data.length())
	{
		//SerialOut("==> decode frq= " + data, true);
GetFreq:
		if (data.indexOf('.'))
		{
			data.remove(data.indexOf('.'), 1);
		}
		uint32_t chk = uint32_t(data.toInt());
		if (chk >= 100000)
		{
			Chk4SyncZeroBeat(chk, false);
			gFreqRadio = chk;
			gFreqRadioPrev = chk;
			ChkFreqModeChng();		
			PrtFreq(false);
			String bnd = CalcBand();
			PrtBand(true);
#ifdef DEBUG_FLEX_API
			SerialOut("Flex set rx freq= " + String(gFreqRadio), true);
#endif
		}

		//chk for new handle for this slice
		if (txt.indexOf(gFlexHandle) == -1 && txt.charAt(0) == 'S')
		{
			//this changes if the flex throws an error, get new handle
			gFlexHandle = txt.substring(0, txt.indexOf('|') + 1);
			gFlexHandle.trim();
			//remove leading 0's - flex doesn't send them in 'S' handle '|'
			while (gFlexHandle.charAt(1) == '0')
			{
				gFlexHandle.remove(1, 1);
			}
			SerialOut("\n\r---> New Handle= " + gFlexHandle + "\n\r", true);			
		}
	}

	//chk for tx frq on this slice
	data = Flex_Extract(gFlexHandle + "transmit freq=", txt);
	if (data.length())
	{
#ifdef DEBUG_FLEX_API
		SerialOut("====> chk tx freq from handle: " + data + ", gFreqRadio= " + String(gFreqRadio), true);
#endif
		goto ChkFrq;
	}

	data = Flex_Extract("|slice " + String(gFlexSlice) + " transmit freq=", txt);
	if (data.length())
	{
#ifdef DEBUG_FLEX_API
		SerialOut("====> chk tx freq from slice: " + txt + ", gRadioFreq= " + String(gFreqRadio), true);
#endif

ChkFrq:
		if (!gFreqRadio)
		{
			goto GetFreq;
		}

		if (data.indexOf('.'))
		{
			data.remove(data.indexOf('.'), 1);
		}
		uint32_t chk = uint32_t(data.toInt());

#ifdef DEBUG_FLEX_API
		SerialOut(String(chk) + ", atu= " + String(gFlexATU) + ", millis()= " + String(millis()) + ", lastTuneMsg= " + String(lastTuneMsg), true);
#endif
		
		if ((chk >= 100000) && (gFlexATU != 2) && ((30000 < millis() - lastTuneMsg) || gFlexCmdNum < 15))
		{
			//if chk is different only change it if we didn't just update it
			if (chk != gFreqRadio && !gFlexLastFreqUpdateTmr)
			{
				//make sure radio stays in sync
#ifdef DEBUG_FLEX_API
				SerialOut("\n\r-->> Setting gFreqRadio to chk " + String(chk) + "\n\r", true);
#endif						
				Chk4SyncZeroBeat(chk, false);
				gFreqRadio = chk;
				gFreqRadioPrev = chk;
				PrtFreq(false);				
			}
		}
	}

	//cw delay
	data = Flex_Extract("break_in_delay=", txt);
	if (data.length())
	{
		gFlexDelay = data.toInt();
	}

	//misc info
	data = Flex_Extract("model=", txt);
	if (data.length())
	{
		//SerialOut("===>model= " + data, true);
	}
	data = Flex_Extract("nickname=", txt);
	if (data.length())
	{
		//SerialOut("===>nickname= " + data, true);
		txt = "";
	}
	data = Flex_Extract("callsign=", txt);
	if (data.length())
	{
		//SerialOut("===>callsign= " + data, true);
	}
		
	//tx info	
	data = Flex_Extract("rfpower=", txt);
	if (data.length())
	{
		//update tx power
		//SerialOut("==> tx power = " + data, true);
		gFlexPO = data.toInt();
	}

	data = Flex_Extract("tunepower=", txt);
	if (data.length())
	{
		//update tx tune power
		//SerialOut("==> tx tune power= " + data, true);
		gFlexTO = data.toInt();
	}

	//atu info
	data = Flex_Extract("atu_enabled=", txt);
	if (data.length())
	{
		//update atu btn status
		lastTuneMsg = millis();
		if (data == "0")
		{
			//SerialOut("-----------> Flex ATU not enabled", true);
			gFlexATU = 0xff; //not installed
		}
		else
		{
			gFlexATU = 0; //exists, standby
			//SerialOut("-----------> Flex ATU enabled", true);
		}
	}
	
	data = Flex_Extract("status=", txt);
	if (data.length())
	{
		//SerialOut("ATU Data: " + data, true);
	}
	if (data == "TUNE_BYPASS" || data == "TUNE_FAIL_BYPASS" || data == "TUNE_MANUAL_BYPASS")
	{
		//SerialOut("==> atu state=" + data, true);

		if (30000 < millis() - lastTuneMsg)
		{
			PrtStatus(data, true);
		}
		lastTuneMsg = millis();
		gFlexATU = 1;
	}
	else if (data == "TUNE_IN_PROGRESS")
	{
		//SerialOut("==> atu state=" + data, true);
		lastTuneMsg = millis();
		PrtStatus(data, true);
		gFlexATU = 2;
	}
	else if (data == "TUNE_SUCCESSFUL")
	{		
		//SerialOut("==> atu state=" + data, true);
		lastTuneMsg = millis();
		PrtStatus(data, true);
		gFlexATU = 3;
	}
}


String Flex_Extract(String match, String txt)
{
	//look for match in txt, if found, extract data value and return with txt empty, otherwise just return empty
	Flex_TxCW(false); //chk for Flex Api keying
	if (!match.length())
	{
		return "";
	}
	String rtnData = "";
	int ptr = txt.indexOf(match);
	if (ptr > -1 && (ptr + match.length()) < txt.length())
	{
		//get data between '=' and ' ' or cr	
		ptr = ptr + match.length();
		int end = txt.indexOf(' ', ptr);
		rtnData = txt.substring(ptr, end); // txt.indexOf(" ", ptr + match.length() + 1));
#if DEBUG_FLEX_API
		SerialOut("\n\r>>>>>" + match + ": " + rtnData + "<<<<<<\n\r", true);	
#endif
	}
	rtnData.trim();
	return rtnData;
}


void Flex_TxFreq(uint32_t freq)
{
	if (gWiFi && gFlexInit)
	{
		//we're online - send freq command	
		String frq = Flex_CalcFreq(gFreqRadio);
		Tx2WiFi(WIFI_FLEX, "|slice t " + String(gFlexSlice) + " " + frq); 
		gFreqRadioPrev = gFreqRadio;
		//SerialOut("Sent " + frq + " to Flex", true);
	}
}


void Flex_TxMode()
{	
	if (gWiFi && gFlexInit)
	{
		String md = Flex_CalcMode();
		
		Tx2WiFi(WIFI_FLEX, "|slice s " + String(gFlexSlice) + " mode=" + md); // +" tx=1 txant=ANT" + String(gFlexAnt));
	}
}

void Flex_RIT()
{

	if (!gRIT)
	{
		Tx2WiFi(WIFI_FLEX, "|slice set " + String(gFlexSlice) + " rit_on=0 rit_freq=0"); 
		//Tx2WiFi(WIFI_FLEX, "|slice set " + String(gFlexSlice) + " rit_freq=0");
	}
	else
	{
		Tx2WiFi(WIFI_FLEX, "|slice set " + String(gFlexSlice) + " rit_on=1 rit_freq=" + String(gRIT));
		//Tx2WiFi(WIFI_FLEX, "|slice set " + String(gFlexSlice) + " rit_freq=" + String(gRIT));
	}
}

void Flex_TxCW(bool forceOff)
{
	static uint32_t ctr = 0;
		
	//if API keying is enabled, send state
	if (gFlexApiKey && gTCPCon && gTxMode)
	{
		if (forceOff || gKeyDown != gKeyDownPrev)
		{
			String txt = "";
			if (gKeyDown && gTxMode && !forceOff)
			{
				Tx2WiFi(WIFI_FLEX, "|cw key 1 time=0x" + String((millis() % 0xffff), HEX) + " index=" + String(ctr) + " client_handle=" + gFlexHandleTx); //send it on TCP once, then UDP three more times to make sure it arrives
				ctr++;
			}
			else
			{
				Tx2WiFi(WIFI_FLEX, "|cw key 0 time=0x" + String((millis() % 0xffff), HEX) + " index=" + String(ctr) + " client_handle=" + gFlexHandleTx); //send it on TCP once, then UDP three more times to make sure it arrives
				ctr++;
			}
			gKeyDownPrev = gKeyDown;
		}
	}

}

void Flex_Update(uint8_t param, uint16_t val, String & txt)
{	
	String id = "";
	switch (param)
	{
	case MNU_FLEX_VOL:
		gFlexVol = val;
		gSaveRadio = true;
		id = gFlexHandle.substring(1, gFlexHandle.length() - 1); //get handle ID#					
		Tx2WiFi(WIFI_FLEX, "|audio client " + id + " slice " + String(gFlexSlice) + " gain " + String(gFlexVol));
		txt = "Updated Audio Level";
		break;
	case MNU_FLEX_PO:
		gFlexPO = val;
		Tx2WiFi(WIFI_FLEX, "|transmit s rfpower=" + String(gFlexPO));
		txt = "Updated Power Out";
		break;
	case MNU_FLEX_TO:
		gFlexTO = val;
		Tx2WiFi(WIFI_FLEX, "|transmit s tunepower=" + String(gFlexTO));
		txt = "Updated Tune Power";
		break;
	case MNU_FLEX_WNB:
		gFlexWNB = val;
		if (gFlexWNB)
		{
			Tx2WiFi(WIFI_FLEX, "|slice s " + String(gFlexSlice) + " wnb=1 wnb_level=" + String(gFlexWNB));
		}
		else
		{
			Tx2WiFi(WIFI_FLEX, "|slice s " + String(gFlexSlice) + " wnb=0");
		}
		txt = "Updated WBand N Blanker";
		break;
	case MNU_FLEX_NB:
		gFlexNB = val;
		if (gFlexNB)
		{
			Tx2WiFi(WIFI_FLEX, "|slice s " + String(gFlexSlice) + " nb=1 nb_level=" + String(gFlexNB));
		}
		else
		{
			Tx2WiFi(WIFI_FLEX, "|slice s " + String(gFlexSlice) + " nb=0");
		}
		txt = "Updated Noise Blanker";
		break;
	case MNU_FLEX_NR:
		gFlexNR = val;
		if (gFlexNR)
		{
			Tx2WiFi(WIFI_FLEX, "|slice s " + String(gFlexSlice) + " nr=1 nr_level=" + String(gFlexNR));
		}
		else
		{
			Tx2WiFi(WIFI_FLEX, "|slice s " + String(gFlexSlice) + " nr=0");
		}
		txt = "Updated Noise Reduction";
		break;
	case MNU_FLEX_AF:
		if (gRMode < RADIO_LSB)
		{
			gFlexAPF = val;
			if (gFlexAPF)
			{
				Tx2WiFi(WIFI_FLEX, "|slice s " + String(gFlexSlice) + " apf=1 apf_level=" + String(gFlexAPF));
			}
			else
			{
				Tx2WiFi(WIFI_FLEX, "|slice s " + String(gFlexSlice) + " apf=0 apf_level=0");
			}
			txt = "Updated Peaking Filter";
		}
		else
		{
			gFlexANF = val;
			if (gFlexANF)
			{
				Tx2WiFi(WIFI_FLEX, "|slice s " + String(gFlexSlice) + " anf=1 anf_level=" + String(gFlexANF));
			}
			else
			{
				Tx2WiFi(WIFI_FLEX, "|slice s " + String(gFlexSlice) + " anf=0 anf_level=0");
			}
			txt = "Updated Notch Filter";
		}
		break;
	case MNU_FLEX_DELAY:
		txt = "Updated Break-in Delay";
		gFlexDelay = val;
		Tx2WiFi(WIFI_FLEX, "|cw break_in_delay " + String(gFlexDelay));
		break;
	}

}