// =====================================================================================================================================================================================================
// (c) 2022 Lynn Hansen, KU7Q															                                                                                                               |
// This Source Code Form is subject to the terms of the GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. A copy of this license can be found here: https://choosealicense.com/licenses/gpl-3.0/|
// =====================================================================================================================================================================================================


//===================================================================================================================
// PCR1000 routines
//===================================================================================================================


uint8_t PCR_Ping()
{
	if ((gBLEConnected || gUDPConnected) && gLinkMiniA && gSyncMode == SYNC_MODE_RMT)
	{
		//we're remote from radio, don't process acutal commands to radio
		if (gTxMode)
		{
			return 2;
		}
		return 0;		
	}

	//request pcr status - if power button is on, return true	
	String reply = "";
	reply = Tx2PCR1000("H1?");	
	if (reply == "H100")
	{		
		gPCRInit = false; //power connected but not turned on	
		return 1;
	}
	else if (reply == "H101")
	{		
		return 2; //power connected and turned on
	}	
	return 0; //power switch truned off
}

void PCR_Initialize()
{
	//initialize PCR1000
	//we just turned on PCR1000, intialize it as described in http://www.thornett.net/Rosliston_Archive_2009-11/ICOM_PCR-1000_Command_Protocol_accessed_July_2011.pdf
	if ((gBLEConnected || gUDPConnected) && gLinkMiniA && gSyncMode == SYNC_MODE_RMT)
	{
		//we're remote from radio, don't process acutal commands to radio
		return;
	}
	
	String reply = "";

	//SerialOut("Initalizing PCR1000", true);
	PCR_Update(MNU_PCR_SQ); //set squelch
	reply = Tx2PCR1000("G300"); //autoupdate off
	reply = Tx2PCR1000("J5100");//disable tone sq
	reply = Tx2PCR1000("J5000"); //disable VSC	
	if (!gPCRHasDSP)
	{
		//ignore if we've previously rcvd a valid dsp reply

		reply = Tx2PCR1000("GD?"); //chk to see if DSP board is installed
		if (reply == "GD01")
		{
			gPCRHasDSP = true;
			SerialOut("PCR1000 has DSP", true);
		}
		else
		{
			gPCRHasDSP = false;
			SerialOut("PCR1000 doesn't have DSP", true);
		}
	}
	
	gPCRDSP = false;
	PCR_Update(MNU_PCR_IFSHIFT); //set IF shift
	PCR_Update(0xff); //set power on/off controls
	PCR_Update(MNU_PCR_DSPNR); //set DSP
	PCR_Update(MNU_PCR_VOL); //set volume
	gPCRInit = true;
	TxFreq2Radio(gFreqRadio);
	delay(50);
	SendLinkParams();
}


void PCR_Update(uint8_t type)
{
	if ((gBLEConnected || gUDPConnected) && gLinkMiniA && gSyncMode == SYNC_MODE_RMT)
	{
		//we're remote from radio, don't process acutal commands to radio
		SendLinkParams();
		return;
	}

	//send new values to the PCR1000 rx based on type	
	String cmd = "";
	String temp = "";
	String reply = "";

	switch (type)
	{
	case 0xff: //power update
		//chk power first. If it changes, just change it and return
		if (gTxMode == RADIO_TX_OFF)
		{
			//turn it off
			reply = Tx2PCR1000("H100");			
			//SerialOut("Turning PCR1000 OFF", true);
		}
		else
		{
			//power now on, initialize radio
			reply = Tx2PCR1000("H101");			
		}
		return;
		break;
	case MNU_PCR_VOL: //update volume				
		cmd = String(gPCRVol, HEX);

		if (cmd.length() == 1)
		{
			cmd = "0" + cmd; //add a leading 0
		}
		cmd = "J40" + cmd;
		//SerialOut("PCR Vol cmd= " + cmd, true);
		reply = Tx2PCR1000(cmd);
		break;
	case MNU_PCR_SQ: //update squelch
		cmd = String(gPCRSq, HEX);
		cmd = String(gPCRSq, HEX);
		if (cmd.length() == 1)
		{
			cmd = "0" + cmd; //add a leading 0
		}
		cmd = "J41" + cmd;
		//SerialOut("PCR Sq cmd= " + cmd, true);
		reply = Tx2PCR1000(cmd);
		break;
	case MNU_PCR_IFSHIFT: //update IF shift
		cmd = String(gPCRIF, HEX);
		if (cmd.length() == 1)
		{
			cmd = "0" + cmd; //add a leading 0
		}
		cmd = "J43" + cmd;
		//SerialOut("PCR IF cmd= " + cmd, true);
		reply = Tx2PCR1000(cmd);
		break;
	case MNU_PCR_DSPNR: //update DSP noise reduction level		
		if (gPCRDSP == 0)
		{
			temp = "0"; //turn off noise reduction
		}
		else
		{
			//valid level 1 to 15
			temp = String(gPCRDSPNR, HEX);
			temp.toUpperCase();
		}
		reply = Tx2PCR1000("G301"); //turn auto update on for the next command
		//DSP commands must be sent as a packet, terminating with CR/LF
		reply = Tx2PCR1000("J8001");
		reply = Tx2PCR1000("J810" + String(gPCRDSP));
		reply = Tx2PCR1000("J820" + temp);
		reply = Tx2PCR1000("J830" + String(gPCRANch));
		reply = Tx2PCR1000("G300"); //turn auto update back off

		break;
	default:
		//something else changed beside power, updated all of the boolean controls
		//SerialOut("Updating ctrls - gPCRStat=" + String(gPCRStat, HEX), true);
		reply = Tx2PCR1000("J450" + String(gPCRAGC));
		reply = Tx2PCR1000("J470" + String(gPCRAtten));
		reply = Tx2PCR1000("J460" + String(gPCRNB));
		break;
	}
}

String Tx2PCR1000(String cmd)
{
	//add cr + lf to cmd, convert string to msg[] and tx it to PCR1000
	//return with reply 		
	bool autoMode = false;
		
	if (cmd == "G301")
	{
		//SerialOut("Auto update on", true);
		autoMode = true; //don't append lf/cr or wait for reply
	}
	else if (cmd == "G300")
	{
		autoMode = false; //back to normal
	}
	

	//clear rx buffer before tx
	int ctr = 0;
	String chr = "";
	while (RADIO.available())
	{		
		chr += char(RADIO.read());
		ctr++;
		//SerialOut(String(chr), false); //do something with it for compiler
	}
	if (ctr > 0)
	{
		//SerialOut("Junk= " + chr + ", ctr= " + String(ctr), true);-->Tx
	}

	byte msg[30];

	//don't add cr/lf for auto mode turn on - it won't reply
	if (autoMode == false)
	{
		cmd = cmd + char(13) + char(10);
	}
	cmd.toUpperCase();
	uint8_t num = cmd.length();
	//SerialOut("-->Tx: " + cmd, false); //don't add cr/lf
	for (uint8_t i = 0; i < num; i++)
	{
		msg[i] = cmd.charAt(i);		
	}		
	RADIO.write(msg, num);		
	if (autoMode == false)
	{		
		return ChkPCR1000Rx(0);
	}
	else
	{
		return " "; //must return with something or audio locks up
	}

}

String ChkPCR1000Rx(uint8_t numChr)
{	
	//wait for PCR1000 to reply, return with reply
	//if numChr > 0, return when that # has been rx
	//if numChr = 0, return when LF has been rx
	uint32_t startTime = millis();
	uint32_t waitTime = 50; //wait up to 50 msec for reply to start - normal time is 7 to 14 mSec
	char chr = 0;
	String msg = "";
	int ctr = 0;

	//SerialOut("Waiting for PCR reply", true);
	while (waitTime > millis() - startTime)
	{		
		if (RADIO.available())
		{			
			if (!ctr)
			{
				//SerialOut("Reply time= " + String(millis() - startTime), true);
			}
			chr = RADIO.read();
			//SerialOut("Incoming - " + String(chr), true);
			if (chr != 13 && chr != 10)
			{
				//ignore lf & cr
				msg += char(chr);
			}
			ctr++;
			startTime = millis(); //reset timer
			if ((numChr != 0 && ctr == numChr) || (numChr == 0 && chr == 10 && ctr > 4))
			{
				if (RADIO.available())
				{
					//more coming, restart
					msg = "";
					ctr = 0;
				}
				else
				{
					//SerialOut("<--Rx: " + msg + "\n", true);
					return msg;
				}
			}
		}
	}
	SerialOut("PCR rx timeout - msg=" + msg,true);
	return " "; //must return with something
}

